@@ -861,33 +861,35 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
         self = env->spare_txns;
         DEBUG("found freelist txn; self=%p self->txn=%p", self, self->txn)
         env->spare_txns = self->spare_next;
-        env->max_spare_txns++;
-        self->flags &= ~TRANS_SPARE;
-        _Py_NewReference((PyObject *)self);
-        UNLOCKED(rc, mdb_txn_renew(self->txn));
 
+        UNLOCKED(rc, mdb_txn_renew(self->txn));
         if(rc) {
             mdb_txn_abort(self->txn);
-            self->txn = NULL;
+            PyObject_Del(self);
+            return err_set("mdb_txn_begin", rc);
         }
+
+        env->max_spare_txns++;
+        self->flags &= ~TRANS_SPARE;
+
     } else {
+        MDB_txn *txn;
         if(write && env->readonly) {
             const char *msg = "Cannot start write transaction with read-only env";
             return err_set(msg, EACCES);
         }
 
-        if(! ((self = PyObject_New(TransObject, &PyTransaction_Type)))) {
-            return NULL;
+        flags = write ? 0 : MDB_RDONLY;
+        UNLOCKED(rc, mdb_txn_begin(env->env, parent_txn, flags, &txn));
+        if(rc) {
+            return err_set("mdb_txn_begin", rc);
         }
 
-        flags = write ? 0 : MDB_RDONLY;
-        UNLOCKED(rc, mdb_txn_begin(env->env, parent_txn, flags, &self->txn));
-    }
-
-    if(rc) {
-        _Py_ForgetReference((PyObject *)self);
-        PyObject_Del(self);
-        return err_set("mdb_txn_begin", rc);
+        if(! ((self = PyObject_New(TransObject, &PyTransaction_Type)))) {
+            mdb_txn_abort(txn);
+            return NULL;
+        }
+        self->txn = txn;
     }
 
     OBJECT_INIT(self)
@@ -917,6 +919,7 @@ static PyObject *
 make_cursor(DbObject *db, TransObject *trans)
 {
     CursorObject *self;
+    MDB_cursor *curs;
     int rc;
 
     if(! trans->valid) {
@@ -928,20 +931,26 @@ make_cursor(DbObject *db, TransObject *trans)
         return NULL;
     }
 
-    self = PyObject_New(CursorObject, &PyCursor_Type);
-    UNLOCKED(rc, mdb_cursor_open(trans->txn, db->dbi, &self->curs));
+    UNLOCKED(rc, mdb_cursor_open(trans->txn, db->dbi, &curs));
     if(rc) {
-        _Py_ForgetReference((PyObject *)self);
-        PyObject_Del(self);
         return err_set("mdb_cursor_open", rc);
     }
 
+    self = PyObject_New(CursorObject, &PyCursor_Type);
+    if (!self) {
+        mdb_cursor_close(curs);
+        return NULL;
+    }
+
     DEBUG("sizeof cursor = %d", (int) sizeof *self)
     OBJECT_INIT(self)
     LINK_CHILD(trans, self)
+    self->curs = curs;
     self->positioned = 0;
     self->key.mv_size = 0;
+    self->key.mv_data = NULL;
     self->val.mv_size = 0;
+    self->val.mv_data = NULL;
     self->trans = trans;
     self->last_mutation = trans->mutations;
     self->dbi_flags = db->flags;
@@ -1041,23 +1050,11 @@ db_flags(DbObject *self, PyObject *args, PyObject *kwds)
     PyObject *dct;
     unsigned int f;
 
-    struct db_flags {
-        TransObject *txn;
-    } arg = {NULL};
-
-    static const struct argspec argspec[] = {
-        {"txn", ARG_TRANS, OFFSET(db_flags, txn)}
-    };
-
-    static PyObject *cache = NULL;
-    if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
-        return NULL;
-    }
-    if(! arg.txn) {
-        return type_error("'txn' argument required");
-    }
-    if(! arg.txn->valid) {
-        return err_invalid();
+    if (args) {
+        Py_ssize_t size = PyTuple_GET_SIZE(args);
+        if(size > 1) {
+            return type_error("too many positional arguments.");
+        }
     }
 
     dct = PyDict_New();
@@ -1370,17 +1367,22 @@ env_copy(EnvObject *self, PyObject *args, PyObject *kwds)
     struct env_copy {
         PyObject *path;
         int compact;
-    } arg = {NULL, 0};
+        TransObject *txn;
+    } arg = {NULL, 0, NULL};
 
     static const struct argspec argspec[] = {
         {"path", ARG_OBJ, OFFSET(env_copy, path)},
-        {"compact", ARG_BOOL, OFFSET(env_copy, compact)}
+        {"compact", ARG_BOOL, OFFSET(env_copy, compact)},
+        {"txn", ARG_TRANS, OFFSET(env_copy, txn)}
     };
 
     PyObject *fspath_obj;
     const char *fspath_s;
     int flags;
     int rc;
+#ifdef HAVE_PATCHED_LMDB
+    MDB_txn *txn;
+#endif
 
     static PyObject *cache = NULL;
     if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
@@ -1393,12 +1395,36 @@ env_copy(EnvObject *self, PyObject *args, PyObject *kwds)
         return NULL;
     }
 
+#ifdef HAVE_PATCHED_LMDB
+    if (arg.txn) {
+        txn = arg.txn->txn;
+        if (!arg.compact) {
+            return type_error("txn argument only compatible with compact=True");
+        }
+    }
+    else {
+        txn = NULL;
+    }
+#else
+    if (arg.txn) {
+        return type_error("Non-patched LMDB doesn't support transaction with env.copy");
+    }
+#endif
+
     fspath_s = PyBytes_AS_STRING(fspath_obj);
     flags = arg.compact ? MDB_CP_COMPACT : 0;
+#ifdef HAVE_PATCHED_LMDB
+    UNLOCKED(rc, mdb_env_copy3(self->env, fspath_s, flags, txn));
+#else
     UNLOCKED(rc, mdb_env_copy2(self->env, fspath_s, flags));
+#endif
     Py_CLEAR(fspath_obj);
     if(rc) {
+#ifdef HAVE_PATCHED_LMDB
+        return err_set("mdb_env_copy3", rc);
+#else
         return err_set("mdb_env_copy2", rc);
+#endif
     }
     Py_RETURN_NONE;
 }
@@ -1412,8 +1438,12 @@ env_copyfd(EnvObject *self, PyObject *args, PyObject *kwds)
     struct env_copyfd {
         int fd;
         int compact;
+        TransObject *txn;
     } arg = {-1, 0};
     int rc;
+#ifdef HAVE_PATCHED_LMDB
+    MDB_txn *txn;
+#endif
 #ifdef _WIN32
     PyObject *temp;
     Py_ssize_t handle;
@@ -1421,7 +1451,8 @@ env_copyfd(EnvObject *self, PyObject *args, PyObject *kwds)
 
     static const struct argspec argspec[] = {
         {"fd", ARG_INT, OFFSET(env_copyfd, fd)},
-        {"compact", ARG_BOOL, OFFSET(env_copyfd, compact)}
+        {"compact", ARG_BOOL, OFFSET(env_copyfd, compact)},
+        {"txn", ARG_TRANS, OFFSET(env_copyfd, txn)},
     };
     int flags;
 
@@ -1434,6 +1465,22 @@ env_copyfd(EnvObject *self, PyObject *args, PyObject *kwds)
     }
     flags = arg.compact ? MDB_CP_COMPACT : 0;
 
+#ifdef HAVE_PATCHED_LMDB
+    if (arg.txn) {
+        txn = arg.txn->txn;
+        if (!arg.compact) {
+            return type_error("txn argument only compatible with compact=True");
+        }
+    }
+    else {
+        txn = NULL;
+    }
+#else
+    if (arg.txn) {
+        return type_error("Non-patched LMDB doesn't support transaction with env.copyfd");
+    }
+#endif
+
 #ifdef _WIN32
     temp = PyObject_CallMethod(msvcrt, "get_osfhandle", "i", arg.fd);
     if(! temp) {
@@ -1444,13 +1491,19 @@ env_copyfd(EnvObject *self, PyObject *args, PyObject *kwds)
     if(PyErr_Occurred()) {
         return NULL;
     }
-    UNLOCKED(rc, mdb_env_copyfd2(self->env, (HANDLE)handle, flags));
+    #define HANDLE_ARG (HANDLE)handle
 #else
-    UNLOCKED(rc, mdb_env_copyfd2(self->env, arg.fd, flags));
+    #define HANDLE_ARG arg.fd
+#endif
+
+#ifdef HAVE_PATCHED_LMDB
+    UNLOCKED(rc, mdb_env_copyfd3(self->env, HANDLE_ARG, flags, txn));
+#else
+    UNLOCKED(rc, mdb_env_copyfd2(self->env, HANDLE_ARG, flags));
 #endif
 
     if(rc) {
-        return err_set("mdb_env_copyfd2", rc);
+        return err_set("mdb_env_copyfd3", rc);
     }
     Py_RETURN_NONE;
 }
@@ -2244,7 +2297,7 @@ cursor_put_multi(CursorObject *self, PyObject *args, PyObject *kwds)
         flags |= MDB_NOOVERWRITE;
     }
     if(arg.append) {
-        flags |= MDB_APPEND;
+        flags |= (self->trans->db->flags & MDB_DUPSORT) ? MDB_APPENDDUP : MDB_APPEND;
     }
 
     if(! ((iter = PyObject_GetIter(arg.items)))) {
@@ -2581,14 +2634,17 @@ static PyObject *
 new_iterator(CursorObject *cursor, IterValFunc val_func, MDB_cursor_op op)
 {
     IterObject *iter = PyObject_New(IterObject, &PyIterator_Type);
-    if(iter) {
-        iter->val_func = val_func;
-        iter->curs = cursor;
-        Py_INCREF(cursor);
-        iter->started = 0;
-        iter->op = op;
+    if (!iter) {
+        return NULL;
     }
-    DEBUG("new_iterator: %#p", (void *)iter)
+
+    iter->val_func = val_func;
+    iter->curs = cursor;
+    Py_INCREF(cursor);
+    iter->started = 0;
+    iter->op = op;
+
+    DEBUG("new_iterator: %p", (void *)iter)
     return (PyObject *) iter;
 }
 
@@ -2992,15 +3048,17 @@ trans_dealloc(TransObject *self)
         self->spare_next = self->env->spare_txns;
         self->env->spare_txns = self;
         self->env->max_spare_txns--;
+        Py_INCREF(self);
 
         Py_CLEAR(self->db);
         UNLINK_CHILD(self->env, self)
         Py_CLEAR(self->env);
-    } else {
-        MDEBUG("deleting trans")
-        trans_clear(self);
-        PyObject_Del(self);
+        return;
     }
+
+    MDEBUG("deleting trans")
+    trans_clear(self);
+    PyObject_Del(self);
 }
 
 /**
@@ -3575,15 +3633,37 @@ enable_drop_gil(void)
  * lmdb.get_version() -> tuple
  */
 static PyObject *
-get_version(void)
+get_version(PyObject *mod, PyObject *args, PyObject *kwds)
 {
+    struct version_args {
+        int subpatch;
+    } arg = {0};
+
+    static const struct argspec argspec[] = {
+        {"subpatch", ARG_BOOL, OFFSET(version_args, subpatch)},
+    };
+
+    static PyObject *cache = NULL;
+    if(parse_args(1, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
+        return NULL;
+    }
+
+    if (arg.subpatch) {
+#ifdef HAVE_PATCHED_LMDB
+        const int subpatch = 1;
+#else
+        const int subpatch = 0;
+#endif
+        return Py_BuildValue("iiii", MDB_VERSION_MAJOR,
+            MDB_VERSION_MINOR, MDB_VERSION_PATCH, subpatch);
+    }
     return Py_BuildValue("iii", MDB_VERSION_MAJOR,
         MDB_VERSION_MINOR, MDB_VERSION_PATCH);
 }
 
 static struct PyMethodDef module_methods[] = {
     {"enable_drop_gil", (PyCFunction) enable_drop_gil, METH_NOARGS, ""},
-    {"version", (PyCFunction) get_version, METH_NOARGS, ""},
+    {"version", (PyCFunction) get_version, METH_VARARGS|METH_KEYWORDS, ""},
     {0, 0, 0, 0}
 };
 
@@ -3618,15 +3698,18 @@ static int init_types(PyObject *mod, PyObject *__all__)
     int i;
     for(i = 0; types[i]; i++) {
         PyTypeObject *type = types[i];
+        char const * name = type->tp_name;
 
         if(PyType_Ready(type)) {
             return -1;
         }
-        if(PyObject_SetAttrString(mod, type->tp_name, (PyObject *)type)) {
+        if(PyObject_SetAttrString(mod, name, (PyObject *)type)) {
             return -1;
         }
 
-        if(type->tp_name[0] != '_' && append_string(__all__, type->tp_name)) {
+        /* As a special case, export _Database */
+        if((name[0] != '_' || 0 == strcmp(name, "_Database")) &&
+           append_string(__all__, name)) {
             return -1;
         }
     }
